import asyncio
import aiosqlite

from typing import Iterable, Any


class Synchronizer:
    def __init__(self):
        self.queue = []

    async def __aenter__(self):
        event = asyncio.Event()
        self.queue.append(event)
        await self.wait_for_opening()

    async def __aexit__(self, *args, **kwargs):
        if len(self.queue) > 1:
            event = self.queue.pop(1)
            event.set()
        else:
            self.queue.pop(0)

    async def wait_for_opening(self) -> None:
        if len(self.queue) < 2:
            return

        await self.queue[-1].wait()



from typing import Any, Union
import json

from .core.message import Message, MessageTemplate
from .core.keyboard import Keyboard


class Dispatcher:
    def __init__(self, event: dict, event_type: str, bot, text: str = None, user_id: int = None, peer_id: int = None,
                 post_id: int = None, owner_id: int = None, object_id: int = None, reply_text: str = None,
                 reply_user_id: int = None, reply_peer_id: int = None, reply_object_id: int = None,
                 action_type: str = None, action_text: str = None, action_object_id: int = None,
                 action_member_id: int = None, payload: dict = None, chain_data: Any = None):
        self.event_type = event_type
        self.bot = bot
        self.user_id = user_id
        self.peer_id = peer_id
        self.chat_id = peer_id - 2000000000
        self.post_id = post_id
        self.owner_id = owner_id
        self.object_id = object_id
        self.event = event
        self.text = text
        self.reply_text = reply_text
        self.reply_user_id = reply_user_id
        self.reply_peer_id = reply_peer_id
        self.reply_object_id = reply_object_id
        self.action_type = action_type
        self.action_text = action_text
        self.action_object_id = action_object_id
        self.action_member_id = action_member_id
        self.payload = payload
        self.chain_data = chain_data

    @property
    def message(self) -> Message:
        return Message(self.bot, self.peer_id, self.object_id)

    @property
    def reply_message(self) -> Message:
        return Message(self.bot, self.peer_id, self.reply_object_id)

    @property
    def forward(self) -> dict:
        return {
            "peer_id": self.peer_id,
            "conversation_message_ids": [self.object_id]
        }

    def create_message_template(self, text: str = None, attachment: str = None,
                                keyboard: Union[str, Keyboard] = None, forward: dict = None) -> MessageTemplate:
        return MessageTemplate(self.bot, text, attachment, keyboard, forward)

    async def answer(self, text: str = None, attachment: str = None, keyboard: Union[str, Keyboard] = None) -> Message:
        response = await self.bot.execute("messages.send", message=text, peer_ids=self.user_id,
                                          random_id=0, attachment=attachment, keyboard=keyboard)
        message_id = response["response"][0]["conversation_message_id"]

        return Message(self.bot, self.peer_id, message_id)

    async def reply(self, text: str = None, attachment: str = None, keyboard: Union[str, Keyboard] = None) -> Message:
        forward = {
            "peer_id": self.peer_id,
            "conversation_message_ids": [self.object_id],
            "is_reply": 1
        }
        forward = json.dumps(forward)

        response = await self.bot.execute("messages.send", message=text, peer_ids=self.peer_id,
                                          random_id=0, attachment=attachment, keyboard=keyboard,
                                          forward=forward)
        message_id = response["response"][0]["conversation_message_id"]

        return Message(self.bot, self.peer_id, message_id)

    async def send_message(self, text: str = None, attachment: str = None, keyboard: Union[str, Keyboard] = None) -> Message:
        response = await self.bot.execute("messages.send", message=text, peer_ids=self.peer_id,
                                          random_id=0, attachment=attachment, keyboard=keyboard)
        message_id = response["response"][0]["conversation_message_id"]

        return Message(self.bot, self.peer_id, message_id)

    async def send_comment(self, text: str = None, attachment: str = None) -> None:
        await self.bot.execute("wall.createComment",
                               owner_id=self.owner_id, post_id=self.post_id, message=text, attachment=attachment)

    async def mark_as_read(self) -> None:
        await self.bot.execute("messages.markAsRead", peer_id=self.peer_id)

    async def set_typing_status(self, typing_status: str = "typing") -> None:
        await self.bot.execute("messages.setActivity", peer_id=self.peer_id, type=typing_status)

    async def kick_user(self, member_id: int) -> None:
        await self.bot.execute("messages.removeChatUser", chat_id=self.chat_id, member_id=member_id)

    async def edit_chat_name(self, title: str) -> None:
        await self.bot.execute("messages.editChat", chat_id=self.chat_id, title=title)


def get_dispatcher_by_event(bot, event: dict, event_params: dict, chain_data: Any) -> Dispatcher:
    event_type = event_params["type"]
    text = event_params["text"]
    user_id = event_params["user_id"]
    peer_id = event_params["peer_id"]
    post_id = event_params["post_id"]
    owner_id = event_params["owner_id"]
    object_id = event_params["object_id"]
    reply = event_params["reply"]
    action = event_params["action"]
    payload = event_params["payload"]

    return Dispatcher(event=event, bot=bot, text=text, user_id=user_id, peer_id=peer_id,
                      post_id=post_id, owner_id=owner_id, reply_text=reply.get("text"),
                      reply_peer_id=reply.get("peer_id"), reply_user_id=reply.get("user_id"),
                      reply_object_id=reply.get("object_id"), chain_data=chain_data, object_id=object_id,
                      action_type=action.get("type"), action_text=action.get("text"), event_type=event_type,
                      action_object_id=action.get("object_id"), action_member_id=action.get("member_id"),
                      payload=payload)



from typing import Union, Callable
from abc import ABC, abstractmethod


class BaseCondition(ABC):
    @abstractmethod
    def new_event(self, event_params: dict) -> bool:
        pass


class Condition(BaseCondition):
    def __init__(self, command: str = None, contains_command: str = None, user_id: int = None,
                 peer_id: int = None, post_id: int = None, owner_id: int = None, functional_condition: Callable = None):
        self.command = command
        self.contains_command = contains_command
        self.user_id = user_id
        self.peer_id = peer_id
        self.post_id = post_id
        self.owner_id = owner_id
        self.functional_condition = functional_condition

    def new_event(self, event_params: dict) -> bool:
        text = event_params["text"]
        user_id = event_params["user_id"]
        peer_id = event_params["peer_id"]
        post_id = event_params["post_id"]
        owner_id = event_params["owner_id"]

        if (self.contains_command is not None and self.contains_command in text) or \
                (self.command is not None and self.command == text) or \
                (self.user_id is not None and self.user_id == user_id) or \
                (self.peer_id is not None and self.peer_id == peer_id) or \
                (self.post_id is not None and self.post_id == post_id) or \
                (self.owner_id is not None and self.owner_id == owner_id) or \
                (self.functional_condition is not None and self.functional_condition(event_params)):
            return True

        return False


class ActionCondition(BaseCondition):
    def __init__(self, action: str = None, command: str = None, member_id: int = None,
                 contains_command: str = None):
        self.action = action
        self.command = command
        self.contains_command = contains_command
        self.member_id = member_id

    def new_event(self, event_params: dict) -> bool:
        text = event_params["action"].get("text")
        member_id = event_params["action"].get("member_id")
        event_type = event_params["action"].get("type")
        print(self.action, event_type)

        if (self.action is not None and self.action == event_type) or \
           (self.contains_command is not None and self.contains_command in text) or \
           (self.command is not None and self.command == text) or \
           (self.member_id is not None and self.member_id == member_id):
            return True

        return False


class PayloadCondition(BaseCondition):
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def new_event(self, event_params: dict) -> bool:
        payload = event_params["payload"]
        for arg, value in self.kwargs.items():
            if payload.get(arg) != value:
                return False

        return True


class And(BaseCondition):
    def __init__(self, *conditions: BaseCondition):
        self.conditions = conditions

    def new_event(self, event_params: dict) -> bool:
        if all(list(map(lambda condition: condition.new_event(event_params), self.conditions))):
            return True

        return False


class Or(BaseCondition):
    def __init__(self, *conditions: BaseCondition):
        self.conditions = conditions

    def new_event(self, event_params: dict) -> bool:
        if any(list(map(lambda condition: condition.new_event(event_params), self.conditions))):
            return True

        return False



from typing import Union, Any
import traceback
import sys

from .dispatcher import get_dispatcher_by_event
from .condition import BaseCondition


class Handler:
    event_type = None

    def __init__(self, condition: BaseCondition = None, is_lower: bool = False, func=None):
        self.condition = condition
        self.is_lower = is_lower
        self.func = func

    def __call__(self, func):
        async def wrapper(bot, event: dict, event_params: dict, chain_data: Any):
            return await func(get_dispatcher_by_event(bot, event, event_params, chain_data))

        return type(self)(self.condition, self.is_lower, wrapper)

    def is_trigger(self, event_params: dict) -> bool:
        if event_params["type"] != self.event_type:
            return False

        if self.is_lower:
            if event_params["text"]:
                event_params["text"] = event_params["text"].lower()
            if "text" in event_params["action"]:
                event_params["action"]["text"] = event_params["action"]["text"].lower()
            if "text" in event_params["reply"]:
                event_params["reply"]["text"] = event_params["reply"]["text"].lower()

        return self.condition is None or self.condition.new_event(event_params)

    async def new_event(self, bot, event: dict, event_params: dict, chain_data: Any = None) -> Any:
        try:
            return await self.func(bot, event, event_params, chain_data)
        except:
            sys.stderr.write(f"\n\n{traceback.format_exc()}\n\n")


class CustomHandler(Handler):
    def __call__(self, func):
        async def wrapper(token: str, event: dict, event_params: dict, chain_data: Any):
            return await func(get_dispatcher_by_event(token, event, event_params, chain_data))

        handler = CustomHandler(self.condition, self.is_lower, wrapper)
        handler.event_type = self.event_type
        return handler


class MessageNewHandler(Handler):
    event_type = "message_new"


class MessageEditHandler(Handler):
    event_type = "message_edit"


class WallReplyNewHandler(Handler):
    event_type = "wall_reply_new"


class WallReplyEditHandler(Handler):
    event_type = "wall_reply_edit"


class WallPostNewHandler(Handler):
    event_type = "wall_post_new"


class BoardPostNewHandler(Handler):
    event_type = "board_post_new"


class BoardPostEditHandler(Handler):
    event_type = "board_post_edit"


class Handlers:
    message_new = MessageNewHandler
    message_edit = MessageEditHandler
    wall_reply_new = WallReplyNewHandler
    wall_reply_edit = WallReplyEditHandler
    wall_post_new = WallPostNewHandler
    board_post_new = BoardPostNewHandler
    board_post_edit = BoardPostEditHandler

    def __new__(cls, event_type):
        handler = CustomHandler()
        handler.event_type = event_type
        return handler



import json


class Button:
    def __init__(self, label: str, color: str, payload: dict = ""):
        self.button = {
            "action": {
                "type": "text",
                "label": label,
                "payload": payload
            },
            "color": color
        }

    def __str__(self):
        return json.dumps(self.button)


class Line:
    def __init__(self, *buttons: Button):
        self.line = []
        for button in buttons:
            self.line.append(button.button)

    def __str__(self):
        return json.dumps(self.line)


class Keyboard:
    def __init__(self, *lines: Line, one_time: bool = False, inline: bool = False):
        self.keyboard = {
            "buttons": []
        }

        if inline:
            self.keyboard["inline"] = inline
        else:
            self.keyboard["one_time"] = one_time

        for line in lines:
            self.keyboard["buttons"].append(line.line)

    def __str__(self):
        return json.dumps(self.keyboard)



import aiohttp
import aiofiles
import json

from .. import Bot


class Uploader:
    def __init__(self, bot: Bot):
        self.bot = bot
        self.dispatcher = None

    async def upload_photo_to_message(self, peer_id: int, path: str):
        response = await self.bot.execute("photos.getMessagesUploadServer", peer_id=peer_id)
        upload_url = response["response"]["upload_url"]

        async with aiofiles.open(path, mode="rb") as file:
            photo_bytes = await file.read()

        file_data = aiohttp.FormData()
        file_data.add_field("photo", photo_bytes)

        async with aiohttp.ClientSession() as session:
            async with session.post(upload_url, data=file_data) as response:
                print(await response.text())
